@@ -6,17 +6,37 @@
  *  gcc -Wall -g rfk.c -o rfk `pkg-config --cflags --libs gtk+-2.0 hildon-1 dbus-glib-1 dbus-1`
  */
 
+/*
+ TO DO:
+  - working demo mode
+  - gtkhtml instructions reader
+*/
+
+#include <dbus/dbus.h>
 #include <dbus/dbus-glib.h>
 #include <gtk/gtk.h>
+#include <libgtkhtml/gtkhtml.h>
 #include <stdlib.h>
 #include <glib.h>
 #include <hildon/hildon.h>
 #include <math.h>
 #include <stdio.h>
 #include <string.h>
+#include <mce/mode-names.h>
+#include <mce/dbus-names.h>
 
-#define ARENA_WIDTH 25
-#define ARENA_HEIGHT 12
+#define MCE_SIGNAL_MATCH "type='signal'," \
+	        "sender='"    MCE_SERVICE     "'," \
+        "path='"      MCE_SIGNAL_PATH "'," \
+        "interface='" MCE_SIGNAL_IF   "'"
+
+#define ARENA_WIDTH 30
+#define ARENA_HEIGHT 16
+
+#define WINDOW_TITLE "robotfindskitten"
+#define WINDOW_TITLE_SHORT "rfk"
+
+#define DEMO_SPEED 200
 
 const int amount_of_random_stuff = 15;
 
@@ -33,14 +53,19 @@ GtkWidget* state_widget[STATE_LAST];
 GSList *nki = NULL;
 guint nki_count = 0;
 
+gboolean portrait_mode = FALSE;
+
 GtkWidget *arena[ARENA_WIDTH][ARENA_HEIGHT];
 GtkWidget *window, *robot, *kitten;
 int robot_x, robot_y;
+int robot_demo_x=-1, robot_demo_y=-1;
 gboolean *used = NULL;
+gboolean demo_running = FALSE;
 
 GdkPixbuf *robot_pic, *love_pic, *kitten_pic;
 
 const GdkColor black = { 0, };
+GdkColor grey = { 0, 0x3333, 0x3333, 0x3333 };
 
 /****************************************************************/
 /* Random object descriptions.                                  */
@@ -125,6 +150,14 @@ random_character (gchar *description)
 /* Talking back to the user.                                    */
 /****************************************************************/
 
+static gboolean
+close_dialogue (gpointer data)
+{
+  GtkDialog *dialogue = (GtkDialog*) data;
+  gtk_dialog_response (dialogue, GTK_RESPONSE_ACCEPT);
+  return FALSE; /* don't go round again */
+}
+
 static void
 show_message (const char *message)
 {
@@ -132,6 +165,10 @@ show_message (const char *message)
     (hildon_note_new_information (GTK_WINDOW (window),
 				  message?message:
 				  "Some message was supposed to be here."));
+  if (demo_running)
+    {
+      g_timeout_add (1000, close_dialogue, note);
+    }
   gtk_dialog_run (GTK_DIALOG (note));
   gtk_widget_destroy (GTK_WIDGET (note));
 }
@@ -150,7 +187,7 @@ ensure_messages_loaded (void)
   if (nki_count)
     return;
 
-  nki_file = fopen ("/usr/share/rfk/non-kitten-items.rfk", "r");
+  nki_file = fopen ("/opt/rfk/non-kitten-items.rfk", "r");
 
   if (!nki_file)
     {
@@ -190,9 +227,9 @@ ensure_messages_loaded (void)
 static void
 load_images (void)
 {
-  robot_pic = gdk_pixbuf_new_from_file ("/usr/share/rfk/rfk-robot.png", NULL);
-  love_pic = gdk_pixbuf_new_from_file ("/usr/share/rfk/rfk-love.png", NULL);
-  kitten_pic = gdk_pixbuf_new_from_file ("/usr/share/rfk/rfk-kitten.png", NULL);
+  robot_pic = gdk_pixbuf_new_from_file ("/opt/rfk/rfk-robot.png", NULL);
+  love_pic = gdk_pixbuf_new_from_file ("/opt/rfk/rfk-love.png", NULL);
+  kitten_pic = gdk_pixbuf_new_from_file ("/opt/rfk/rfk-kitten.png", NULL);
 }
 
 /****************************************************************/
@@ -216,7 +253,8 @@ switch_state (StateOfPlay new_state)
 }
 
 /****************************************************************/
-/* Things we need DBus for: online help, and vibration.         */
+/* Things we need DBus for: online help, vibration, and         */
+/* orientation.                                                 */
 /****************************************************************/
 
 static void
@@ -243,26 +281,70 @@ call_dbus (DBusBusType type,
 
   proxy = dbus_g_proxy_new_for_name (connection, name, path, interface);
 
-  error = NULL;
-  if (!dbus_g_proxy_call (proxy, method, &error,
-			  G_TYPE_STRING, parameter,
-			  G_TYPE_INVALID,
-			  G_TYPE_INVALID))
-    {
-      show_message (error->message);
-      g_error_free (error);
-    }
+  dbus_g_proxy_call_no_reply (proxy, method,
+		  G_TYPE_STRING, parameter,
+		  G_TYPE_INVALID);
+}
+
+/*
+ * Online help, such as it is.
+ */
+static void
+help_supply_image (gpointer html_dummy,
+		const char *url,
+		HtmlStream *stream,
+		gpointer data_dummy)
+{
+	gchar *img;
+	guint size;
+	
+     if (g_file_get_contents (url,
+			     &img, &size, NULL))
+       {
+	  html_stream_write (stream, img, size);
+	  g_free (img);
+       }
+
+     html_stream_close (stream);
 }
 
 static gboolean
 get_help (gpointer button, gpointer data)
 {
-  call_dbus (DBUS_BUS_SESSION,
-	     "com.nokia.osso_browser",
-	     "/com/nokia/osso_browser/request",
-	     "com.nokia.osso_browser",
-	     "load_url",
-	     "/usr/share/rfk/help.html");
+  GtkWidget *window = hildon_stackable_window_new ();
+  static gchar *contents = NULL;
+  static HtmlDocument *doc = NULL;
+  GtkWidget *pan = NULL;
+  GtkWidget *html = NULL;
+
+  if (!contents &&
+      !g_file_get_contents ("/opt/rfk/help.html",
+        &contents,
+        NULL, NULL))
+    {
+      show_message ("Sorry, the help could not be displayed.");
+      g_free (window);
+      return FALSE;
+    }
+
+  if (!doc)
+    {
+      doc = html_document_new ();
+      g_signal_connect (G_OBJECT (doc), "request-url", G_CALLBACK (help_supply_image), NULL);
+      html_document_open_stream (doc, "text/html");
+      html_document_write_stream (doc, contents, -1);
+      html_document_close_stream (doc);
+    }
+
+  html = html_view_new ();
+  g_signal_connect (G_OBJECT (html), "request-url", G_CALLBACK (help_supply_image), NULL);
+  g_object_set(G_OBJECT(html), "sensitive", FALSE, NULL);
+  html_view_set_document (HTML_VIEW (html), doc);
+  pan = hildon_pannable_area_new ();
+  gtk_container_add (GTK_CONTAINER (pan), html);
+  gtk_container_add (GTK_CONTAINER (window), pan);
+  gtk_window_set_title (GTK_WINDOW (window), WINDOW_TITLE " help");
+  gtk_widget_show_all (window);
   return FALSE;
 }
 
@@ -277,6 +359,73 @@ vibrate (void)
 	     "PatternIncomingMessage");
 }
 
+
+static DBusHandlerResult
+mce_filter_func (DBusConnection * connection,
+                 DBusMessage * message, gpointer dummy)
+{
+  DBusError error;
+  char *rotation, *stand, *face;
+  int x, y, z;
+	
+  if (!dbus_message_is_signal (message, MCE_SIGNAL_IF, MCE_DEVICE_ORIENTATION_SIG))
+    return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+
+  dbus_error_init (&error);
+  if (dbus_message_get_args (message,
+                             &error,
+                             DBUS_TYPE_STRING, &rotation,
+                             DBUS_TYPE_STRING, &stand,
+                             DBUS_TYPE_STRING, &face,
+                             DBUS_TYPE_INT32,  &x,
+                             DBUS_TYPE_INT32,  &y,
+                             DBUS_TYPE_INT32,  &z, DBUS_TYPE_INVALID))
+    {
+      /* Rotate main window */
+      if (strcmp (rotation, MCE_ORIENTATION_PORTRAIT)==0)
+        {		
+          portrait_mode = TRUE;
+          hildon_gtk_window_set_portrait_flags (GTK_WINDOW (window), HILDON_PORTRAIT_MODE_REQUEST|HILDON_PORTRAIT_MODE_SUPPORT);
+	  gtk_window_set_title (GTK_WINDOW (window), WINDOW_TITLE_SHORT);
+        }
+      else
+        {
+          portrait_mode = FALSE;
+          hildon_gtk_window_set_portrait_flags (GTK_WINDOW (window), HILDON_PORTRAIT_MODE_SUPPORT);
+	  gtk_window_set_title (GTK_WINDOW (window), WINDOW_TITLE);
+        }
+    }
+  else
+    {
+      g_warning ("%s: %s\n", error.name, error.message);
+      dbus_error_free (&error);
+    }
+
+  return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+}	    
+
+static void
+check_for_rotation (void)
+{
+  DBusConnection *connection = dbus_bus_get (DBUS_BUS_SYSTEM, NULL);
+  
+  call_dbus (DBUS_BUS_SYSTEM,
+	     MCE_SERVICE,
+	     MCE_REQUEST_PATH,
+	     MCE_REQUEST_IF,
+	     "req_accelerometer_enable",
+	     NULL);
+
+  dbus_bus_add_match (connection, MCE_SIGNAL_MATCH, NULL);
+
+  dbus_connection_add_filter (connection,
+                               (DBusHandleMessageFunction) mce_filter_func,
+                               NULL, NULL);
+  
+  hildon_gtk_window_set_portrait_flags (GTK_WINDOW (window),
+                                        HILDON_PORTRAIT_MODE_SUPPORT);
+}
+
 /****************************************************************/
 /* The ending animation.                                        */
 /****************************************************************/
@@ -434,8 +583,10 @@ static gboolean
 move_robot (guint8 whichway)
 {
   GtkWidget *new_space;
-  gint8 dx = directions[whichway].move_x;
-  gint8 dy = directions[whichway].move_y;
+  gint8 dx, dy;
+  
+  dx = directions[whichway].move_x;
+  dy = directions[whichway].move_y;
 
   const char *found;
 
@@ -457,6 +608,8 @@ move_robot (guint8 whichway)
 	  switch_state (STATE_EPILOGUE);
 	}
 
+      robot_demo_x = robot_demo_y = -1;
+      g_object_set_data (G_OBJECT (new_space), "visited", "");
       return TRUE;
     }
   else
@@ -477,6 +630,109 @@ move_robot (guint8 whichway)
     }
 }
 
+static void
+move_robot_randomly (void)
+{
+  guint8 whichway;
+  gint8 dx, dy;
+  guint8 i;
+
+  for (i=0; i<100; i++)
+    {
+      whichway = random() % 8;
+      dx = directions[whichway].move_x;
+      dy = directions[whichway].move_y;
+
+      /* Can't move off the edge. */
+      if (robot_x+dx<0 ||
+	  robot_y+dy<0 ||
+	  robot_x+dx>=ARENA_WIDTH ||
+	  robot_y+dy>=ARENA_HEIGHT)
+	continue;
+
+      /* Can't move onto an occupied square randomly. */
+      if (g_object_get_data (G_OBJECT (arena[robot_x+dx][robot_y+dy]),
+			     "examine"))
+	continue;
+
+      /* Success! */
+      move_robot (whichway);
+      return;
+    }
+
+  /* if we get here, robot is stuck or just very unlucky;
+   * either way we do nothing.
+   */
+}
+
+static gboolean
+item_is_investigatable (guint8 x, guint8 y)
+{
+  GObject *item = G_OBJECT (arena[x][y]);
+
+  /* FIXME make sure this doesn't trigger on robot! */
+  if (g_object_get_data (item, "examine")==NULL)
+    /* Empty space */
+    return FALSE;
+
+  if (g_object_get_data (item, "visited")!=NULL)
+    /* Been there, done that */
+    return FALSE;
+
+  return TRUE;
+}
+
+static gboolean
+move_robot_demo (gpointer dummy)
+{
+  gint8 whichway;
+  gint8 dx, dy;
+
+  if (current_state!=STATE_PLAYING)
+    return FALSE;
+
+  if (robot_demo_x == -1)
+    {
+      /* Find an item to investigate. */
+
+      robot_demo_x = robot_demo_y = 0;
+
+      while (!item_is_investigatable (robot_demo_x,
+				      robot_demo_y))
+	{
+	  robot_demo_x++;
+	  if (robot_demo_x==ARENA_WIDTH)
+	    {
+	      robot_demo_x=0;
+	      robot_demo_y++;
+	      if (robot_demo_y==ARENA_HEIGHT)
+		{
+		  robot_demo_y=0;
+		}
+	    }
+	}
+    }
+  
+  whichway = ((int) (7+atan2 (robot_x-robot_demo_x, robot_y-robot_demo_y)*4/M_PI)) % 8;
+  dx = directions[whichway].move_x;
+  dy = directions[whichway].move_y;
+
+  if (!(robot_x+dx == robot_demo_x && robot_y+dy == robot_demo_y)
+      &&
+      g_object_get_data (G_OBJECT (arena[robot_x+dx][robot_y+dy]),
+			 "examine"))
+    {
+      /* Can't move onto an occupied square, unless it's the target. */
+      move_robot_randomly ();
+    }
+  else
+    {
+      move_robot (whichway);
+    }
+
+  return TRUE;
+}
+
 /****************************************************************/
 /* Event handlers.                                              */
 /****************************************************************/
@@ -486,23 +742,36 @@ on_window_clicked (GtkWidget      *widget,
 		   GdkEventButton *event,
 		   gpointer        user_data)
 {
-  /** Centre point of robot's representation on screen */
-  int rx, ry;
-  double angle;
+  int quarter_width, quarter_height;
+  int quadrant;
+  int directions[] = { 0, 7, 6, 1, -1, 5, 2, 3, 4 };
 
   if (current_state!=STATE_PLAYING)
     {
       return TRUE;
     }
 
-  rx = (robot->allocation.x+robot->allocation.width/2);
-  ry = (robot->allocation.y+robot->allocation.height/2);
+  quarter_width = window->allocation.width/4;
+  quarter_height = window->allocation.height/4;
 
-  angle = atan2(event->x - rx,
-		event->y - ry) +
-    M_PI * (9/8);
+  if (event->x < quarter_width)
+    quadrant = 0;
+  else if (event->x > quarter_width*3)
+    quadrant = 2;
+  else
+    quadrant = 1;
 
-  move_robot (((int) (angle / (M_PI/4))) % 8);
+  if (event->y < quarter_height)
+    /* nothing */;
+  else if (event->y > quarter_height*3)
+    quadrant += 6;
+  else
+    quadrant += 3;
+
+  quadrant = directions[quadrant];
+
+  if (quadrant >= 0)
+    move_robot (quadrant);
 
   return TRUE;
 }
@@ -530,26 +799,39 @@ on_key_pressed (GtkWidget      *widget,
       if (keyval==directions[i].gdk_key ||
 	  keyval==directions[i].vi_key)
 	{
+          int whichway = i;
+
+	  if (portrait_mode)
+             whichway = (whichway+6)%8;
+
 	  if (event->state & GDK_SHIFT_MASK)
 	    {
-	      while (!move_robot (i))
+	      while (!move_robot (whichway))
 		{
 		  /* keep going, robot! */
 		}
 	    }
 	  else
 	    {
-	      move_robot (i);
+	      move_robot (whichway);
 	    }
 	  return FALSE;
 	}
     }
 
-  if (keyval=='d' && event->state & GDK_CONTROL_MASK)
+  if (keyval=='q' && event->state & GDK_CONTROL_MASK)
     {
       /* secret debugging key */
       show_message (gtk_label_get_text (GTK_LABEL (kitten)));
     }
+  else if (keyval=='r')
+    {
+      move_robot_randomly ();
+    }
+  else if (keyval=='d')
+    {
+      move_robot_demo (NULL);
+    }
 
   return FALSE;
 }
@@ -557,6 +839,15 @@ on_key_pressed (GtkWidget      *widget,
 static void
 play_game (gpointer button, gpointer data)
 {
+  demo_running = FALSE;
+  switch_state (STATE_PLAYING);
+}
+
+static void
+play_demo (gpointer button, gpointer data)
+{
+  demo_running = TRUE;
+  g_timeout_add (DEMO_SPEED, move_robot_demo, NULL);
   switch_state (STATE_PLAYING);
 }
 
@@ -593,6 +884,7 @@ set_up_board (void)
 
       g_object_unref (robot);
       g_object_unref (kitten);
+      robot_demo_x = robot_demo_y = -1;
     }
   else
     {
@@ -602,6 +894,8 @@ set_up_board (void)
       used = g_malloc0 (nki_count * sizeof(gboolean));
 
       robot = gtk_label_new ("#");
+      /* make sure robot doesn't try to visit itself */
+      g_object_set_data (G_OBJECT (robot), "visited", "");
       g_object_ref (robot);
       kitten = random_character ("You found kitten!  Way to go, robot!");
       g_object_ref (kitten);
@@ -626,6 +920,38 @@ set_up_board (void)
     }
 }
 
+static gboolean
+backdrop_draw (GtkWidget *widget, GdkEventExpose *event, gpointer data)
+{
+  GdkGC *gc = NULL;
+  int quarter_width = event->area.width / 4;
+  int quarter_height = event->area.height / 4;
+
+  gc = gdk_gc_new (GDK_DRAWABLE (widget->window));
+  gdk_gc_set_foreground (gc, &grey);
+  gdk_draw_rectangle (GDK_DRAWABLE (widget->window),
+		  gc,
+		  TRUE,
+		  quarter_width, 0,
+		  quarter_width*2, quarter_height);
+   gdk_draw_rectangle (GDK_DRAWABLE (widget->window),
+		  gc,
+		  TRUE,
+		  quarter_width, quarter_height*3,
+		  quarter_width*2, quarter_height);
+   gdk_draw_rectangle (GDK_DRAWABLE (widget->window),
+		  gc,
+		  TRUE,
+		  0, quarter_height,
+		  quarter_width, quarter_height*2);
+   gdk_draw_rectangle (GDK_DRAWABLE (widget->window),
+		  gc,
+		  TRUE,
+		  quarter_width*3, quarter_height,
+		  quarter_width, quarter_height*2);
+  return FALSE;
+}
+
 static void
 set_up_widgets (void)
 {
@@ -638,7 +964,7 @@ set_up_widgets (void)
     "by the existence of various things which are not kitten. "
     "Robot must touch items to determine if they are kitten or "
     "not. The game ends when robotfindskitten. You may move "
-    "robot about by tapping on any side of robot, or with the "
+    "robot about by tapping on any side of the screen, or with the "
     "arrow keys.";
   GKeyFile *desktop = g_key_file_new ();
   gchar *version;
@@ -647,13 +973,15 @@ set_up_widgets (void)
   
   /* The window */
 
-  window = hildon_window_new ();
-  gtk_window_set_title (GTK_WINDOW (window), "robotfindskitten");
+  window = hildon_stackable_window_new ();
+  gtk_window_set_title (GTK_WINDOW (window), WINDOW_TITLE);
   gtk_widget_modify_bg (window, GTK_STATE_NORMAL, &black);
   g_signal_connect (G_OBJECT (window), "button-press-event", G_CALLBACK (on_window_clicked), NULL);
   g_signal_connect (G_OBJECT (window), "key-press-event", G_CALLBACK (on_key_pressed), NULL);
   g_signal_connect (G_OBJECT (window), "delete_event", G_CALLBACK (gtk_main_quit), NULL);
 
+  gdk_colormap_alloc_color (gdk_colormap_get_system (), &grey, TRUE, TRUE);
+
   /* The prologue */
 
   /* Get the rather odd version string.  The RFK spec says that
@@ -678,7 +1006,13 @@ set_up_widgets (void)
 					     HILDON_BUTTON_ARRANGEMENT_HORIZONTAL,
 					     "Play", NULL);
   g_signal_connect (button, "clicked", G_CALLBACK (play_game), NULL);
+  gtk_box_pack_end (GTK_BOX (buttons), button, TRUE, TRUE, 0);
 
+
+  button = hildon_button_new_with_text (HILDON_SIZE_AUTO_WIDTH | HILDON_SIZE_THUMB_HEIGHT,
+					     HILDON_BUTTON_ARRANGEMENT_HORIZONTAL,
+					     "Demo", NULL);
+  g_signal_connect (button, "clicked", G_CALLBACK (play_demo), NULL);
   gtk_box_pack_end (GTK_BOX (buttons), button, TRUE, TRUE, 0);
 
   button = hildon_button_new_with_text (HILDON_SIZE_AUTO_WIDTH | HILDON_SIZE_THUMB_HEIGHT,
@@ -702,6 +1036,7 @@ set_up_widgets (void)
   explain = gtk_label_new (explanation);
   gtk_label_set_line_wrap (GTK_LABEL (explain), TRUE);
 
+  gtk_box_pack_end (GTK_BOX (middle), gtk_image_new_from_pixbuf (kitten_pic), FALSE, FALSE, 0);
   gtk_box_pack_end (GTK_BOX (middle), explain, TRUE, TRUE, 0);
   gtk_box_pack_end (GTK_BOX (middle), gtk_image_new_from_pixbuf (robot_pic), FALSE, FALSE, 0);
 
@@ -714,13 +1049,16 @@ set_up_widgets (void)
   state_widget[STATE_PROLOGUE] = intro;
 
   /* The game itself */
-
   state_widget[STATE_PLAYING] = gtk_table_new (ARENA_HEIGHT, ARENA_WIDTH, TRUE);
   g_signal_connect (state_widget[STATE_PLAYING], "parent-set", G_CALLBACK (set_up_board), NULL);
 
   for (x=0; x < ARENA_WIDTH; x++)
     for (y=0; y < ARENA_HEIGHT; y++)
       arena[x][y] = NULL;
+  
+  g_signal_connect (G_OBJECT (state_widget[STATE_PLAYING]),
+		  "expose_event",
+		  G_CALLBACK (backdrop_draw), NULL);
 
   /* The epilogue */
   state_widget[STATE_EPILOGUE] =  gtk_drawing_area_new ();
@@ -751,6 +1089,7 @@ main (gint argc,
   load_images ();
 
   set_up_widgets ();
+  check_for_rotation ();
   switch_state (STATE_PROLOGUE);
   
   gtk_main ();
